/*
   hpt.c  ... 'Element tests with hypoplsticity'
   
     Program for computation of oedometer and triaxial tests
     with the hypoplastic constitutive equation
     (see e.g. von Wolffersdorff, Mechanics of Cohesive-Frictional Materials,
               1(4),1996,251-271)
     
     I. Herle, Karlsruhe, 30.01.1997
     
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define pi 3.14159265

#define DEBUG
#undef DEBUG

double T[4];     /* stress vector (t1,t2,t3)             */
double D[4];     /* stretching rate  (d1,d2,d3)          */
double Tk[4];    /* stress rate                          */
double E[4];     /* vector of deformation                */
double e;        /* void ratio                           */

struct hparam {  /* hypoplastc parameters                */
  double phic;   /* critical friction angle [deg]        */
  double hs;     /* granular hardness  [MPa]             */
  double n;      /* exponent                             */
  double ed0;    /* minimum void ratio at p=0            */
  double ec0;    /* critical void ratio at p=0           */
  double ei0;    /* maximum void ratio at p=0            */
  double alpha;  /* strength exponent                    */
  double beta;   /* stiffness exponent                   */
} par ;

double a, hi;    /* parameters of phic and ex0           */
double dt;       /* time step                            */


/* -------------------------------------------------------------- */

double deg2rad(alpha_deg)
/* Converts from degrees into radians */
double alpha_deg;
{
  return alpha_deg*pi/180.;
}

/* -------------------------------------------------------------- */

double tr3(A)
double A[];
{
  return A[1]+A[2]+A[3];
}

double norm3(A)
double A[];
{
  return sqrt(A[1]*A[1]+A[2]*A[2]+A[3]*A[3]); 
}

void readln0(fname)
/* read a line without information */
FILE *fname;
{
  char c;
  
  while ( (c = getc(fname)) != '\n' ) {;}
}

void tx(fd,F)
/* computation of the unknown stretching rate in TX-test after Niemunis */
double fd,F;
{
  double L21,L22,N2;
  double xfak,trT,aa,b,c,diskr,x1,x2,x;
  
  trT = tr3(T);
  L21 = a*a*T[1]*T[2]/trT/trT;
  L22 = (F*F+2.*a*a*T[2]*T[2]/trT/trT);
  N2 = F*fd*a/3.*(4.*T[2]-T[1])/trT;
  aa = 2.*N2*N2-L22*L22;  /* (a ist globale Variable !) */
  b = 4.*L21*N2*D[1];
  c = (L22*L22+2.*L21*L21)*D[1]*D[1];
  diskr = b*b-4.*aa*c;
  if (diskr<0) {
    printf("No real solution exists! \n");
    exit (-3); }
  else {
    x1 = (-b+sqrt(diskr))/2./aa;
    x2 = (-b-sqrt(diskr))/2./aa;
  }
  if (x1<0 && x2<0) {
    printf("No real solution exists! \n");
    exit (-3); }
  else if (x1>0 && x2>0) {
    printf("Solution is not unique! \n");
    exit (-3); }
  else {
    if (x1>0) 
      x = x1;
    else
      x = x2;
  }
  D[2] = -L21/L22*D[1]-N2/L22*x;
  D[3] = D[2];
}

void stg_gw(ttyp)
/* constitutive equation for the stress rate */
char ttyp;
{
  double  fe, fb, fd;
  double Td[4], Td2[4], Tsd[4], Tsd2[4], Tsd3[4], TdD[4];
  double ei, ec, ed, tan_psi, cos_3th, F;
  double expon;
  int i;
  
  for(i=1; i<=3; i++) {
    Td[i] = T[i]/tr3(T);
    Td2[i] = Td[i]*Td[i];
    Tsd[i] = Td[i]-1./3.;
    Tsd2[i] = Tsd[i]*Tsd[i];
    Tsd3[i] = Tsd2[i]*Tsd[i];
  }
  
  expon = exp(-pow(-tr3(T)/par.hs,par.n));
  ei = par.ei0*expon;
  ec = par.ec0*expon;
  ed = par.ed0*expon;
  
  fe = pow(ec/e,par.beta);
  fd = pow((e-ed)/(ec-ed),par.alpha);
  fb = hi*(1.+ei)/ei*pow(-tr3(T)/par.hs,1.-par.n);
  
  tan_psi = sqrt(3.)*norm3(Tsd);
  if (tan_psi == 0.)
    F = 1.;
  else {
    cos_3th = -sqrt(6.)*tr3(Tsd3)/pow(tr3(Tsd2),1.5);
    if (cos_3th>1) 
      cos_3th = 1.;
    else if (cos_3th<-1)
      cos_3th = -1.;
    F = sqrt( tan_psi*tan_psi/8.+(2.-tan_psi*tan_psi) /
                 (2.+sqrt(2.)*tan_psi*cos_3th) ) - tan_psi/2./sqrt(2.);
  }

  /* unknown stress rate in TXD-Versuch */
  if ( ttyp == 't' ) tx(fd,F);
  
  for(i=1; i<=3; i++) TdD[i] = Td[i]*D[i];

  for(i=1; i<=3; i++) {
    Tk[i] = fb*fe/tr3(Td2)* 
            (F*F*D[i]+a*a*tr3(TdD)*Td[i]+fd*a*F*(Td[i]+Tsd[i])*norm3(D));
  }
}

struct grenze { double T1; double E1; };

double limits(max,richtung,dt)
/* are the test limits already reached?
    yes --> correction of the last dt to reach the limit exactly */
struct grenze max;
char richtung;
double dt;
{
  double dtl=0.;
  
  if (richtung=='1') {    /* compression */
    if ( (T[1]+Tk[1]*dt)<max.T1 )
      dtl = (max.T1-T[1])/Tk[1]; 
    else if ( (E[1]+D[1]*dt)<max.E1 ) 
      dtl = (max.E1-E[1])/D[1];
      } 
  else if (richtung=='2') {   /* extension */
    if ( (T[1]+Tk[1]*dt)>max.T1 )
      dtl = (T[1]-max.T1)/Tk[1]; 
    else if ( (E[1]+D[1]*dt)>max.E1 ) 
      dtl = (E[1]-max.E1)/D[1];
  }
  return dtl; 
}


/* -------------------------------------------------------------- */

main()
{
  struct grenze {
    double T1;  /* maximum stress T1 */
    double E1;  /* maximum strain E1 */
  } max;
  double dtl; /* time step for the last increment */
  int i, j=1;
  char ttyp, richtung;  /* ttyp ... test type:
                               ='o'...oedometer
                               ='i'...isotropic compression
                               ='t'...standard triaxial (drained)
                               ='u'...triaxial with c.v. (undrained)
                            richtung
                               ='1'...compression
                               ='2'...extension    */
  FILE *fout, *fin, *fhp;

/* open INPUT- and OUTPUT files */
  if ((fin = fopen("hp_inp.par","r"))==NULL) {
    printf("File 'hp_inp.par' cannot be opened \n");
    return (-1);
  }
  if ((fhp = fopen("hp_stg.par","r"))==NULL) {
    printf("File 'hp_stg.par' cannot be opened \n");
    return (-1);
  }
  if ((fout = fopen("hp_out.dat","w+"))==NULL) {
    printf("File 'hp_out.dat' cannot be opened \n");
    return (-1);
  }
  
/* Read initial values */
  readln0(fin); readln0(fin); readln0(fin);   /* 3 Kopfzeilen */
  /* type of test */
  fscanf(fin,"%c %c",&ttyp,&richtung);
  readln0(fin); readln0(fin);
  /* initial state (stress, void ratio) */
  fscanf(fin,"%lf %lf",&T[1],&T[2]);
    T[1] = -T[1];
    T[2] = -T[2];
  readln0(fin); readln0(fin);  
  fscanf(fin,"%lf",&e);
  /* test limits */
  readln0(fin); readln0(fin);  
  fscanf(fin,"%lf %lf",&max.T1,&max.E1);
    max.T1 = -max.T1;
    max.E1 = -max.E1;
  fclose(fin);
/* material parameters*/
  readln0(fhp); readln0(fhp);   /* 2 heading lines */
  fscanf(fhp,"%lf %lf %lf",&par.phic,&par.hs,&par.n);
  readln0(fhp); readln0(fhp);  
  fscanf(fhp,"%lf %lf %lf",&par.ed0,&par.ec0,&par.ei0);
  readln0(fhp); readln0(fhp);  
  fscanf(fhp,"%lf %lf",&par.alpha,&par.beta);
  fclose(fhp);

/* test of INPUT data */
  if ( ttyp!='o' && ttyp!='t' && ttyp!='u' && ttyp!='i' ) {
    printf("False test type! \n");
    return (-2); }
  else if ( richtung!='1' && richtung!='2' ) {
    printf("False loading direction! \n");
    return (-2); }
  else if ( T[1]>=0 || T[2]>=0 ) {
    printf("Pressure must be positive! \n");
    return (-2); }
  else if ( e<=0 ) {
    printf("Void ratio must be positive! \n");
    return (-2); }
  else if ( max.T1>=0 ) {
    printf("False test limits! \n");
    return (-2); }
  else if ( par.phic<=0 || par.hs<=0 || par.ed0<=0 || par.ec0<=0 ||
            par.ei0<=0 || par.alpha<0 || par.beta<0 || par.n<0) {
    printf("Uncorrect material parameters! \n");
    return (-2); }
  else if ( par.ed0>par.ec0 || par.ed0>par.ei0 || par.ec0>par.ei0 ) {
    printf("It must be e_i0>e_c0>e_d0 ! \n");
    return (-2); 
  }

/* Initial values */
  T[3] = T[2];
  if (richtung=='1')
    D[1] = -1.0;
  else if (richtung=='2')
    D[1] = 1.0;
  if (ttyp=='u') { D[2]=-D[1]*0.5; D[3]=D[2]; }
  if (ttyp=='i') { D[2]=D[1]; D[3]=D[1]; }
  dt = 0.0005;
    
/* parameters a and hi */
  a = sqrt(3.)*(3.-sin(deg2rad(par.phic))) / 
      (2.*sqrt(2.)*sin(deg2rad(par.phic)));
  hi = par.hs/par.n*pow(par.ei0/par.ec0,par.beta) /
      ( 3.+a*a-sqrt(3.)*a*pow((par.ei0-par.ed0)/(par.ec0-par.ed0),par.alpha) );

/* computation loop -- start */
  do {
  
    /* output */
    fprintf(fout,"%6lf  %6lf  %6lf  %6lf  %5lf \n",T[1],T[2],E[1],E[2],e);
    
    /* stress rate */
    stg_gw(ttyp);
    
    /* limits */
    dtl = limits(max,richtung,dt);
    if (dtl) dt = dtl;
    
    /* integration */
    for(i=1; i<=3; i++) {
      T[i] += Tk[i]*dt;
      E[i] += D[i]*dt;
    }  
    e += (1.+e)*tr3(D)*dt;
    
#ifdef DEBUG
    j++; if(j>10) break;
#endif
    
  } while( !dtl );
/* computation loop -- end */

/* last step */
  fprintf(fout,"%6lf  %6lf  %6lf  %6lf  %5lf \n",T[1],T[2],E[1],E[2],e);
  
  fclose(fout);
}
