//----------------------------------------------------------------------------
//  Project hypoplastic model
//  
//  Copyright  1997 - 1999. All Rights Reserved.
//
//  SUBSYSTEM:    hypo.app Application
//  FILE:         wolfferdoc.h
//  AUTHOR:       Doanh T. based on hyptest.c of I. Herle
//
//  OVERVIEW
//  ~~~~~~~~
//  Class definition for QHypoplasticModel, QWolffersdorffModel,
//		QHypoplasticDoc (TFileDocument).
//
//----------------------------------------------------------------------------
#if !defined(HWOLFFER_HYPO_DOC_H)              // Sentry, use file only if it's not already included.
#define HWOLFFER_HYPO_DOC_H

#if !defined(__MATH_H)
	#include <math.h>
#endif
#if !defined(__CLASSLIB_ARRAYS_H)
	#include <classlib\arrays.h>
#endif

#include "OWL_doanh.h"
#pragma hdrstop

#if !defined(OWL_DC_H)
	#include <owl/dc.h>
#endif
#if !defined(OWL_DOCMANAG_H)
	#include <owl/docmanag.h>
#endif
#if !defined(OWL_FILEDOC_H)
	#include <owl/filedoc.h>
#endif


const int vnParametersChanged = vnCustomBase+0;
NOTIFY_SIG(vnParametersChanged,void)

#define EV_VN_CHANGTPARAMETRES  VN_DEFINE(vnParametersChanged,  VnChangtparam,void)

typedef enum {
   UNKNOWN,
	WOLFFERSDORFF,
  	WOLFFERSDORFF_WU,
   WEI_WU,
   KOLYMBAS,
   WOLFFERSDORFF_4PARAM} hModel;

typedef enum {
	TRIAXIAL_DRAINED,
   TRIAXIAL_UNDRAINED,
   ISOTROPIC,
   OEDOMETRIC,
   LINEAR_STRESS_PATH,
   LINEAR_STRAIN_PATH,
   STRAIN_RESPONSE_ENVELOPE,
  	STRESS_RESPONSE_ENVELOPE,
   CIRCULAR_STRESS_PATH,
   CIRCULAR_STAIN_PATH} hTestType;

   	// COMPRESSION is not accepted by compiler !
typedef enum {
	D_COMPRESSION,
   D_EXTENSION} hTestDirection;

struct HMaterialParameter{ // material hypoplastic parameters /
	double phi; 					// critical frictional angle (degree)
	double hs; 		 			// granular hardness (MPA)
	double n; 					// exponent
	double ed0; 					// minimum void ratio at P=0
	double ec0; 					// critical void ratio at P=0
	double ei0; 					// maximum void ratio at P=0
	double alpha; 				// strength exponent
	double beta; 				// stiffness exponent
};

struct HTestParameter{			// test simulation parameters
	hTestType test_type; 		// O,I,T,U
	hTestDirection direction; 	// 1=compression, 2= extension
	double dt; 						// time step integration
  	double defdIncAngle; 			// example, of CKD test
	double sigmamax;  				// stop condition of simulation
	double epsilonmax;
};

struct HMaterialState{ //material state
	// 4 elements, index 0 not used, --> option base one: 1,2,3
	double sigma[4]; 		// stress vector
	double Dsigma[4]; 		// stress increment (delta sigma)
	double epsilon[4]; 	// strain
	double Depsilon[4]; 	// strain increment (delta epsilon)
	double e; 				// void ratio
};

/*  ***************************************************************
one hypoplastic result
	 for use with BIDS
	 with operator == and default constructor
******************************************************************* */
class HHypoResults{ 	// one point of hypoplastic result

public:
	HHypoResults(){
      sigma1 = sigma2 = sigma3 = epsilon1 = epsilon2
      		 = epsilon3 = e = 0.0;
	}
	double sigma1;     	// stress 3 dimensions
  	double sigma2;
	double sigma3;
	double epsilon1;  		// strain 3 dimensions
  	double epsilon2;
	double epsilon3;
	double e;         		// void ratio

   void operator =(const HHypoResults& other)
	{
		sigma1 = other.sigma1;
    	sigma2 = other.sigma2;
    	sigma3 = other.sigma3;
    	epsilon1 = other.epsilon1;
    	epsilon2 = other.epsilon2;
      epsilon3 = other.epsilon3;
    	e = other.e;
	}

	bool operator ==(const HHypoResults& other) const
	{
		return &other == this;
	}
	friend ostream& operator <<(ostream& os, HHypoResults& doc){
      char c = ',';
		os << doc.sigma1 << c << doc.sigma2 << c << doc.sigma3 << c
    		<<	doc.epsilon1 << c << doc.epsilon2 << c << doc.epsilon3 << c
         << doc.e << endl;
		return os;
	}
	friend istream& operator >>(istream& is, HHypoResults& doc){
		char c;
    	is >> doc.sigma1 >> c >> doc.sigma2 >> c >> doc.sigma3 >> c
    	 	>> doc.epsilon1 >> c >> doc.epsilon2 >> c >> doc.epsilon3 >> c
       	>> doc.e;
    	is.ignore();
		return is;
	}

};

typedef TArray<HHypoResults> HHypoResultsVector;
typedef TArrayIterator<HHypoResults> HHypoResultsVectorIterator;
typedef TIArrayAsVector<HHypoResults> IAHHypoResults;
typedef TIArrayAsVectorIterator<HHypoResults> IAHHypoResultsIterator;

/*  ***************************************************************
Abstract class for hypoplastic model
******************************************************************* */
class QHypoplasticModel {
public:

	QHypoplasticModel():fSimulatedData(10,0,10){}
	virtual ~QHypoplasticModel(){}

		// virtual methods of hypoplastic simulation
	virtual void HInitValues() = 0;
	virtual void HInitSimulation() = 0;
	virtual int HSimulation() = 0;
   virtual int HIntermediateConstants(void) = 0;
	virtual void HMaterialChanged(HMaterialParameter&) = 0;
	virtual void HTestChanged(HTestParameter&) = 0;
	virtual void HInitialStateChanged(HHypoResults&) = 0;

			// in public section for views and parameter dialog
	HHypoResultsVector 	fSimulatedData;
	HMaterialParameter 	fMaterialParameter;
	HTestParameter 		fTestParameter;
	HMaterialState 		fCurrentState;

   	// default values of model
	double defphi, defhs, defn, defed0, defec0, defei0, defalpha, defbeta;     // model parameters
	double defe, defsigma[4], defepsilon[4], defDsigma[4], defDepsilon[4];	  // initial state
  	double defdIncAngle;                                                       // constant stress or strain increment
	double defdt, defsigmamax, defepsilonmax;                             	  // stop condition
	hTestType deftype;                                              		  	  // test type
   hTestDirection defdir;                                              		  // test direction
   hModel defmodel;                                              		  		  // model
  	int 	error;                                                              // simulation error


protected:
	virtual int CoreModel(void) = 0;
	virtual int TriaxStrainIncControlled(void) = 0;
  	virtual int HFindStrainIncrement(void) = 0;
  	virtual int HLinearStressPath(double fromX1, double fromY1, double toX2, double toY2) = 0;
   virtual int HCircularStressPath(double fromSigma3, double fromSigma1, double radius) = 0;
   virtual int HLinearStrainPath(void) = 0;
  	virtual int ClassicalTests(void) = 0;
  	virtual int SpecialTests(void) = 0;
	virtual double Limits(void) = 0;

   virtual int HStrainResponseEnvelope(double fromSigma3, double fromSigma1) = 0;
   virtual int HStressResponseEnvelope(double fromSigma3, double fromSigma1) = 0;

	static double Trace3(double A[]){
		return A[1]+A[2]+A[3];}
	static double Norm3(double A[]){
		return sqrt(A[1]*A[1]+A[2]*A[2]+A[3]*A[3]);}
	static double Deg2Rad(double angle){
		return angle*M_PI/180.0;}
  	static double Rad2Deg(double rad){
		return rad*180.0/M_PI;}
};

/*  ***************************************************************
hypoplastic model of Wolffersdorff (comprehensive model)

   .. hypoplastic model of P.A. von Wolffersdorff
   "A hypoplastic relation for granular materials with a predefined limit state surface"
   Mechanics of Cohesive-Frictional Materials, vol 1, 4, 251 - 271, 1996
   .. adaptation of hyp.c of I.HERLE, 30.01.1997
   for oedometer and triaxial tests

******************************************************************* */
class QWolffersdorffModel : public QHypoplasticModel {

public:

	QWolffersdorffModel(){defmodel = WOLFFERSDORFF;}
	virtual ~QWolffersdorffModel(){}

		// virtual methods of QHypoplasticModel
	virtual void HInitValues();
	virtual void HInitSimulation();
	virtual int HSimulation();
	virtual void HMaterialChanged(HMaterialParameter&);
	virtual void HTestChanged(HTestParameter&);
	virtual void HInitialStateChanged(HHypoResults&);

protected:
	virtual int CoreModel(void);
	virtual int TriaxStrainIncControlled(void);
  	virtual int HFindStrainIncrement(void);
  	virtual int HLinearStressPath(double fromX1, double fromY1, double toX2, double toY2);
   virtual int HCircularStressPath(double fromSigma3, double fromSigma1, double radius);
   virtual int HLinearStrainPath(void);
  	virtual int ClassicalTests(void);
  	virtual int SpecialTests(void);
	virtual double Limits(void);

   virtual int HStrainResponseEnvelope(double fromSigma3, double fromSigma1);
   virtual int HStressResponseEnvelope(double fromSigma3, double fromSigma1);

   	// utilities
   virtual int HIntermediateConstants(void);
   int HParameterAandHi();
   void HSetCurrentState(HHypoResults& currentResults);
   int HCheckFromStrainState(void);

   virtual int HSolve(void);

   		// Intermediate Constants
   double f_a, f_hi, f_ei, f_ec, f_ed, f_fb, f_fd, f_fe, f_F;
};

/*  ***************************************************************
extending the hypoplastic model of Wolffersdorff by W. Wu

   .. hypoplastic model of P.A. von Wolffersdorff
   "A hypoplastic relation for granular materials with a predefined limit state surface"
   Mechanics of Cohesive-Frictional Materials, vol 1, 4, 251 - 271, 1996
   .. adaptation of hyp.c of I.HERLE, 30.01.1997
   for oedometer and triaxial tests

******************************************************************* */
class QWolffersdorffExtendedModel : public QWolffersdorffModel {
public:

	QWolffersdorffExtendedModel():QWolffersdorffModel(){defmodel = WOLFFERSDORFF_WU;}
	virtual ~QWolffersdorffExtendedModel(){}

protected:
	//virtual int CoreModel(double a, double hi);
  	virtual int HFindStrainIncrement(void);

   virtual int HSolve(void);
};

class QWeiWuModel : public QWolffersdorffModel {
public:

	QWeiWuModel():QWolffersdorffModel(){defmodel = WEI_WU;}
	virtual ~QWeiWuModel(){}

   double fC1, fC2, fC3, fC4;     // model parameters

   virtual void HInitValues();
   virtual void HInitSimulation();
protected:
	//virtual int CoreModel(double a, double hi);
  	virtual int HFindStrainIncrement(void);

   virtual int HSolve(void);

   void HFindCoefC(void);
   virtual void HFindCoefCi(double *T, double *dT1, double *D1, double *dT2, double *D2, double *C);
};

class QKolymbasModel : public QWeiWuModel {
public:

	QKolymbasModel():QWeiWuModel(){defmodel = KOLYMBAS;}
	virtual ~QKolymbasModel(){}

   virtual void HInitValues();
protected:
	//virtual int CoreModel(double a, double hi);
  	virtual int HFindStrainIncrement(void);

   virtual int HSolve(void);

   virtual void HFindCoefCi(double *T, double *dT1, double *D1, double *dT2, double *D2, double *C);
};

class QWolffersdorff4ParamModel : public QWeiWuModel {
public:

	QWolffersdorff4ParamModel():QWeiWuModel(){defmodel = WOLFFERSDORFF_4PARAM;}
	virtual ~QWolffersdorff4ParamModel(){}

   virtual void HInitValues();
protected:
	//virtual int CoreModel(double a, double hi);
  	virtual int HFindStrainIncrement(void);

   virtual int HSolve(void);

   virtual void HFindCoefCi(double *T, double *dT1, double *D1, double *dT2, double *D2, double *C);
};

/*  ***************************************************************
hypoplastic document
******************************************************************* */
class _USERCLASS QHypoplasticDoc: public TFileDocument {
public:
	enum {
		PrevProperty = TFileDocument::NextProperty-1,
		DocumentClass,
		TEmplateName,
		Viewcount,
		StoragePath,
		DocTitle,
		NextProperty,
	 };

	QHypoplasticDoc(TDocument* parent = 0);
	~QHypoplasticDoc();

 		// implement virtual methods of TDocument
	bool   Open(int mode, const char far* path=0);
	bool   Close(){return TRUE;}
	bool   IsOpen(){
		return fHypoModel->fSimulatedData.GetItemsInContainer()!=0;
	}

	bool   Commit(bool force=false);
	bool   Revert(bool clear=false)
  	{
		if (!TFileDocument::Revert(clear))
			return false;
		if (!clear)
			Open(0);
			return true;
		}

//	 int    FindProperty(const char far* name);  // return index
//	 int    PropertyFlags(int index);
//	 const char* PropertyName(int index);
//	 int    PropertyCount()
//	 {
//		return NextProperty - 1;
//	 }
//	 int    GetProperty(int index, void far* dest, int textlen=0);
protected:
	string FileInfo;

		// experimental comparison
public:
	void HShowExperimentalData();
	void HRemoveExperimentalData();

  	void HMaterialChanged(HMaterialParameter&);
	void HTestChanged(HTestParameter&);
	void HInitialStateChanged(HHypoResults&);
   short HInputParameters(TWindow* aWindow);

   int HGetTestDataFromStream(istream& is);
   int HGetDocDataFromStream(istream& is);

			// in public section for views and parameter dialog
   QHypoplasticModel 	*fHypoModel;
	HHypoResultsVector 	*fExperimentalData;
   HHypoResultsVector 	*fExperimentalDataCopy;

};

#endif  //  sentry.
/********************* EOF ************************************************/
